c     Copyright (C) 2008-2009 Vincent Eymet
c
c     Planet_EMC is free software; you can redistribute it and/or modify
c     it under the terms of the GNU General Public License as published by
c     the Free Software Foundation; either version 3, or (at your option)
c     any later version.

c     Planet_EMC is distributed in the hope that it will be useful,
c     but WITHOUT ANY WARRANTY; without even the implied warranty of
c     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c     GNU General Public License for more details.

c     You should have received a copy of the GNU General Public License
c     along with Planet_EMC; if not, see <http://www.gnu.org/licenses/>
c
      subroutine create_cumulative(Nelem,elem,cumul)
      implicit none
      include 'max.inc'
c
c     Purpose: to create a cumulative from discrete values
c
c     Inputs:
c       + Nelem: number of discrete values
c       + elem: array of 'Nelem' discrete values
c
c     Output:
c       + cumul: cumulative (0-1)
c

c     Inputs
      integer Nelem
      double precision elem(1:Ncumul_mx)
c     Outputs
      double precision cumul(0:Ncumul_mx)
c     temp
      integer strlen,i
      double precision sum
      character*100 label
      label='subroutine create_cumulative'

      if (Nelem.gt.Ncumul_mx) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Nelem:',Nelem
         write(*,*) 'is greater than Ncumul_mx=',Ncumul_mx
         stop
      endif

      cumul(0)=0.0D+0
      do i=1,Nelem
         cumul(i)=cumul(i-1)+elem(i)
      enddo
      sum=cumul(Nelem)
      do i=1,Nelem
         cumul(i)=cumul(i)/sum
      enddo

      return
      end


      subroutine invert_cumulative(Nelem,cumul,idx)
      implicit none
      include 'max.inc'
c
c     Purpose: to invert the provided (discrete) cumulative and
c     return the index of the chosen element
c
c     Inputs:
c       + Nelem: size of discrete cumulative array
c       + cumul: discrete cumulative array
c
c     Output:
c       + idx: chosen element
c

c     Inputs
      integer Nelem
      double precision cumul(0:Ncumul_mx)
c     Ouput
      integer idx
c     temp
      integer i,idxf,strlen
      double precision R
      character*100 label
      label='subroutine invert_cumulative'

      call random_gen(R)

      idxf=0
      do i=1,Nelem
         if ((R.gt.cumul(i-1)).and.(R.le.cumul(i))) then
            idx=i
            idxf=1
            goto 122
         endif
      enddo
 122  continue
      if (idxf.eq.0) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Value of R could not be inverted'
         write(*,*) 'R=',R
         write(*,*) 'Nelem=',Nelem
         do i=1,Nelem
            write(*,*) 'cumul(',i,')=',cumul(i)
         enddo
         stop
      endif

      return
      end
