c     Copyright (C) 2008-2009 Vincent Eymet
c
c     Planet_EMC is free software; you can redistribute it and/or modify
c     it under the terms of the GNU General Public License as published by
c     the Free Software Foundation; either version 3, or (at your option)
c     any later version.

c     Planet_EMC is distributed in the hope that it will be useful,
c     but WITHOUT ANY WARRANTY; without even the implied warranty of
c     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c     GNU General Public License for more details.

c     You should have received a copy of the GNU General Public License
c     along with Planet_EMC; if not, see <http://www.gnu.org/licenses/>
c
      subroutine last_chance(dim,event,Nr,Ntheta,Nphi,r,theta,phi,
     &     rspace,r_nodes,r_south_pole,r_north_pole,
     &     coeff_a,coeff_b,coeff_c,P_in,u_in,
     &     ir_out,itheta_out,iphi_out,P_out,u_out,s_out,valid)
      implicit none
      include 'mpif.h'
      include 'max.inc'
      include 'param.inc'
c
c     Purpose: to give a last chance to the few photons that are definitely "lost" into
c     the geometry;
c     This routine will be called as a last solution, when all other strategies failed.
c     Typically, this is for photons that are in a different gas cell they actually should,
c     or when path exits the current gas cell too soon.
c
c     Inputs
c       + dim: dimension of the vectors
c       + event: index of the event
c       + Nr,Ntheta,Nphi: number of cells
c       + rspace: radius at space level
c       + r, theta, phi: limits of the spherical cells
c       + r_nodes: radius (altitude) at each one of the 4 nodes for
c                  each (theta,phi) angular sector
c       + r_south_pole: radius at south pole
c       + r_north_pole: radius at north pole
c       + coeff_a, coeff_b, coeff_c: a, b and c coefficients for each quarter
c       + P_in: (r,theta,phi) spherical coordinates of the entry point
c       + u_in: entry direction
c
c     Outputs
c       + ir_out,itheta_out,iphi_out: index of the cell the particle really enters
c       + P_out: position (possibly modified) of the particle when it enters the cell
c       + u_out: propagation direction of the particle when it enters the cell
c       + s_out: surface index
c       + valid: 1: a solution was found; 0 otherwise
c

c     inputs
      integer dim,event
      integer Nr(1:Ntheta_mx,1:Nphi_mx),Ntheta,Nphi
      double precision r(0:Nz_mx,1:Ntheta_mx,1:Nphi_mx)
      double precision theta(0:Ntheta_mx)
      double precision phi(0:Nphi_mx)
      double precision rspace
      double precision r_nodes(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision r_south_pole,r_north_pole
      double precision coeff_a(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision coeff_b(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision coeff_c(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision P_in(1:Nmat_mx)
      double precision u_in(1:Nmat_mx)
c     outputs
      integer ir_out,itheta_out,iphi_out
      double precision P_out(1:Nmat_mx)
      double precision u_out(1:Nmat_mx)
      integer s_out,valid
c     MPICH
      integer ierr,errorcode,np,pindex
      integer stat(MPI_STATUS_SIZE)
c     MPICH
c     temp
      integer ioe,edge
      double precision pos(1:2),r_ground
      double precision dr_min
      double precision ca(1:4)
      double precision cb(1:4)
      double precision cc(1:4)
      integer cellf,surface_int,cf
      integer ir,itheta,iphi
      integer ir1,itheta1,iphi1
      double precision P(1:Nmat_mx)
      double precision P1(1:Nmat_mx)
      double precision Pint(1:Nmat_mx)
      double precision Pint1(1:Nmat_mx)
      double precision u(1:Nmat_mx)
      double precision u1(1:Nmat_mx)
      double precision n(1:Nmat_mx),prod
      double precision length
      integer i,surface,app,solf,aerr
      integer solf1,aerr1,surface1
      integer surface_out
      integer niter,nit,Nit_max,it,s
      integer senst,sensp
      double precision thetac,phic
      double precision dtheta,dphi
      character*100 log_file
      character*100 string
      character*100 stringp_file
      integer nl
      integer strlen
      character*100 label
c     MPICH
c     Get 'np', the number of processes
      call MPI_COMM_SIZE(MPI_COMM_WORLD,np,ierr)
c     Get 'pindex', the index of the current process
      call MPI_COMM_RANK(MPI_COMM_WORLD,pindex,ierr)
c     MPICH
      label='subroutine last_chance'
      dtheta=1.0D-8 ! rad
      dphi=1.0D-8 ! rad

      call logfile_name(pindex,event,log_file)
      string='./nlines'
      call stringp_file_name(pindex,string,stringp_file)
      call get_nlines(log_file,stringp_file,nl)

      open(10,file=log_file(1:strlen(log_file)))
c     Debug
c      write(*,*) 'nl1=',nl
c     Debug
      do i=1,nl
         read(10,*)
      enddo
c     Debug
c      write(*,*) 'ok1'
c     Debug
      if (nl.gt.0) then
         write(10,*)
      endif

      write(10,*) 'This is routine ',label(1:strlen(label))
c     Debug
      write(10,*) 'P_in=',(P_in(i),i=1,dim)
      write(10,*) 'u_in=',(u_in(i),i=1,dim)
      close(10)
c     Debug

      Nit_max=50
      valid=0
      niter=0
      nit=0
      call find_dr_min(Nr,Ntheta,Nphi,r,dr_min)
      length=dr_min/1.5D+0             ! m
      s=0

      call copy_vector(dim,P_in,P)
      call copy_vector(dim,u_in,u)
      call scalar_vector(dim,-1.0D+0,u,u1)

 111  continue
      it=0
 112  continue
c     identify the cell where "P" is on the surface, and "u" is an input direction
      call identify_valid_cell(dim,event,Nr,Ntheta,Nphi,r,theta,phi,
     &     r_nodes,r_south_pole,r_north_pole,
     &     coeff_a,coeff_b,coeff_c,P,u,
     &     cellf,ir,itheta,iphi,surface,n)
c     Debug

      call logfile_name(pindex,event,log_file)
      string='./nlines'
      call stringp_file_name(pindex,string,stringp_file)
      call get_nlines(log_file,stringp_file,nl)
      open(10,file=log_file(1:strlen(log_file)))
c     Debug
c      write(*,*) 'nl2=',nl
c     Debug
      do i=1,nl
         read(10,*)
      enddo
c     Debug
c      write(*,*) 'ok2'
c     Debug
      if (nl.gt.0) then
         write(10,*)
      endif

      write(10,*) 'cellf=',cellf
      if (cellf.eq.1) then
         write(10,*) 'ir=',ir,' Nr=',Nr(itheta,iphi)
         write(10,*) 'r=',r(ir-1,itheta,iphi),
     &        r(ir,itheta,iphi)
         write(10,*) 'itheta=',itheta
         write(10,*) 'theta=',theta(itheta-1),theta(itheta)
         write(10,*) 'iphi=',iphi
         write(10,*) 'phi=',phi(iphi-1),phi(iphi)
      endif
c      call extract_coeffs(itheta,iphi,coeff_a,coeff_b,coeff_c,
c     &     ca,cb,cc)
c      pos(1)=P_in(2)
c      pos(2)=P_in(3)
c      call ground_radius(pos,theta(itheta-1),theta(itheta),
c     &     phi(iphi-1),phi(iphi),ca,cb,cc,r_ground)
c      write(10,*) 'r_ground=',r_ground
c     Debug
      if (cellf.eq.0) then
c     give it a last shot, in the case phi=0 or 2*pi
         if ((P(3).ge.2.0D+0*pi).and.(it.eq.0)) then
            it=it+1
            P(3)=0.0D+0
            s=1
            write(10,*) 'Trying again using P(3)=',P(3)
            close(10)
            goto 112
         endif
         if ((P(3).le.0.0D+0).and.(it.eq.0)) then
            it=it+1
            P(3)=2.0D+0*pi
            s=-1
            write(10,*) 'Trying again using P(3)=',P(3)
            close(10)
            goto 112
         endif
         goto 666
      endif
c     Debug
c      write(10,*) 'ir=',ir,' Nr=',Nr(itheta,iphi)
c      write(10,*) 'r=',r(ir-1,itheta,iphi),
c     &     r(ir,itheta,iphi)
c      write(10,*) 'itheta=',itheta
c      write(10,*) 'theta=',theta(itheta-1),theta(itheta)
c      write(10,*) 'iphi=',iphi
c      write(10,*) 'phi=',phi(iphi-1),phi(iphi)
c     Debug

c     case when valid cell was found, but with an iteration
      if ((cellf.eq.1).and.(it.eq.1)) then
         if (s.eq.1) then
            P_in(3)=0.0D+0
c         write(10,*) 'P_in(3) was set to zero'
         else if (s.eq.-1) then
            P_in(3)=2.0D+0*pi
         else
            write(10,*) 'Error from ',
     &           label(1:strlen(label)),' :'
            write(10,*) 's=',s
            stop
         endif
      endif

c     Debug
      write(10,*) 'now extract_coeffs with:'
      write(10,*) 'itheta=',itheta
      write(10,*) 'iphi=',iphi
c     Debug

      call extract_coeffs(itheta,iphi,coeff_a,coeff_b,coeff_c,
     &     ca,cb,cc)
      app=0
      if (ir.eq.1) then
         call need2approximate(dim,Ntheta,r,theta,phi,
     &        r_nodes,r_south_pole,r_north_pole,itheta,iphi,app)
      endif
c     Debug
      write(10,*) 'ca=',(ca(i),i=1,4)
      write(10,*) 'cb=',(cb(i),i=1,4)
      write(10,*) 'cc=',(cc(i),i=1,4)
      write(10,*) 'app=',app
c     Debug

c     case of particles that enter on a cell edge at space level
      if (P(1).eq.rspace) then
         call is_on_edge(P,r(Nr(itheta,iphi)-1,itheta,iphi),
     &        r(Nr(itheta,iphi),itheta,iphi),
     &        theta(itheta-1),theta(itheta),
     &        phi(iphi-1),phi(iphi),
     &        app,ca,cb,cc,ioe,edge)
         if (ioe.eq.1) then
c            write(10,*) 'particle enter through edge:',edge
            call jump_to(dim,P_in,u_in,1.0D+2,P1)
c            write(10,*) 'P1=',(P1(i),i=1,dim)
            call identify_cell(dim,P1,Nr,Ntheta,Nphi,r,theta,phi,
     &           coeff_a,coeff_b,coeff_c,
     &           r_nodes,r_south_pole,r_north_pole,
     &           cf,ir1,itheta1,iphi1)
            if (cf.eq.0) then
               write(10,*) 'particle enters through edge at space level'
               write(10,*) 'cell could not be identified for P1'
               valid=0
               goto 666
c               stop
            else
c               write(10,*) 'valid entry cell:'
c               write(10,*) 'ir1=',ir1
c               write(10,*) 'r=',r(ir1-1,itheta1,iphi1),
c     &              r(ir1,itheta1,iphi1)
c               write(10,*) 'itheta=',itheta1
c               write(10,*) 'theta=',theta(itheta1-1),theta(itheta1)
c               write(10,*) 'iphi=',iphi1
c               write(10,*) 'phi=',phi(iphi1-1),phi(iphi1)               
               valid=1
               ir_out=ir1
               itheta_out=itheta1
               iphi_out=iphi1
               call copy_vector(dim,P,P_out)
               call copy_vector(dim,u,u_out)
               s_out=2
               goto 666
            endif
c         else
c            write(10,*) 'particle does not enter through cell edge'
         endif ! is on edge
      endif ! P is at space level

c     other particles that enter through and edge, but not at space level
c      write(10,*) 'P=',(P(i),i=1,dim)
      call is_on_edge(P,r(ir-1,itheta,iphi),
     &     r(ir,itheta,iphi),
     &     theta(itheta-1),theta(itheta),
     &     phi(iphi-1),phi(iphi),
     &     app,ca,cb,cc,ioe,edge)
      if (ioe.eq.1) then
         call jump_to(dim,P_in,u_in,length,P1)
         call identify_cell(dim,P1,Nr,Ntheta,Nphi,r,theta,phi,
     &        coeff_a,coeff_b,coeff_c,
     &        r_nodes,r_south_pole,r_north_pole,
     &        cf,ir,itheta,iphi)
         if (cf.eq.0) then
            write(10,*) 'particle enters through edge'
            write(10,*) 'cell could not be identified for P1'
            valid=0
            goto 666
c            stop
         endif
      endif

c     Debug
      write(10,*) 'now intersection with:'
      write(10,*) 'P=',(P(i),i=1,dim)
      write(10,*) 'u=',(u(i),i=1,dim)
      write(10,*) 'app=',app
      close(10)
c     Debug

c     once a valid cell has been found, test intersections with this cell
      call intersection(0,dim,event,app,1,1,P,u,
     &     r(ir-1,itheta,iphi),r(ir,itheta,iphi),
     &     theta(itheta-1),theta(itheta),
     &     phi(iphi-1),phi(iphi),ca,cb,cc,1,1,
     &     solf,surface_int,Pint,aerr)
c     Debug
      call logfile_name(pindex,event,log_file)
      string='./nlines'
      call stringp_file_name(pindex,string,stringp_file)
      call get_nlines(log_file,stringp_file,nl)
      open(10,file=log_file(1:strlen(log_file)))
c     Debug
c      write(*,*) 'nl3=',nl
c     Debug
      do i=1,nl
         read(10,*)
      enddo
c     Debug
c      write(*,*) 'ok3'
c     Debug
      if (nl.gt.0) then
         write(10,*)
      endif
      write(10,*) 'solf=',solf,' aerr=',aerr
      if ((solf.eq.1).and.(aerr.eq.0)) then
         write(10,*) 'Pint=',(Pint(i),i=1,dim)
         write(10,*) 'surface_int=',surface_int
      endif
c     Debug
      if ((solf.eq.0).or.(aerr.eq.1)) then
 222     continue
c     Debug
         write(10,*) 'No intersection found with cell'
c     Debug
         if (niter.lt.Nit_max) then
            niter=niter+1
            nit=nit+1

            if (niter.eq.1) then
               if (P(2).lt.thetac) then
                  senst=1
               else
                  senst=-1
               endif
               if (P(3).lt.phic) then
                  sensp=1
               else
                  sensp=-1
               endif
            endif

            write(10,*) 'trying again, iteration:',niter,' /',Nit_max
            if (((ir.eq.1).and.(iphi.eq.1).and.(u(1).le.0.0D+0).and.
     &           (u(2).le.0.0D+0)).or.
     &           ((ir.eq.1).and.(iphi.eq.Nphi).and.(u(1).le.0.0D+0)
     &           .and.(u(2).ge.0.0D+0))) then
c     VERY VERY VERY special case of a particle trying to exit both at phi=0 and r=r_ground
               if (iphi.eq.1) then
                  P(3)=0.1D+0 ! rad
               else if (iphi.eq.Nphi) then
                  P(3)=2.0D+0*pi-0.1D+0 ! rad
               endif
               close(10)
               goto 111
            endif
            call jump_to(dim,P_in,u_in,length*dble(nit),P1)
c     Debug
            write(10,*) 'Position P1=',(P1(i),i=1,dim)
c     Debug
            call identify_cell(dim,P1,Nr,Ntheta,Nphi,r,theta,phi,
     &           coeff_a,coeff_b,coeff_c,
     &           r_nodes,r_south_pole,r_north_pole,
     &           cf,ir1,itheta1,iphi1)
            if (cf.eq.0) then
c     Debug
               write(10,*) 'Error from ',
     &              label(1:strlen(label)),' :'
               write(10,*) 'could not find a valid cell'
               write(10,*) 'Position P1=',(P1(i),i=1,dim)
c               goto 666

c     Trying again with a slightly different input position
               thetac=(theta(itheta-1)+theta(itheta))/2.0D+0
               phic=(phi(iphi-1)+phi(iphi))/2.0D+0
               P(2)=P(2)+dtheta*dble(senst)
               P(3)=P(3)+dphi*dble(sensp)

               nit=0
               write(10,*) 'Now trying with modified position:'
               write(10,*) 'P=',(P(i),i=1,dim)
               close(10)
               goto 111
            else
               write(10,*) 'Position P1=',(P1(i),i=1,dim)
               write(10,*) 'was identified in cell:',ir1,itheta1,iphi1
c     Debug
            endif
            call extract_coeffs(itheta1,iphi1,coeff_a,coeff_b,coeff_c,
     &           ca,cb,cc)
            app=0
            if (ir.eq.1) then
               call need2approximate(dim,Ntheta,r,theta,phi,
     &              r_nodes,r_south_pole,r_north_pole,itheta1,iphi1,app)
            endif
c     Debug
            write(10,*) 'ca=',(ca(i),i=1,4)
            write(10,*) 'cb=',(cb(i),i=1,4)
            write(10,*) 'cc=',(cc(i),i=1,4)
            write(10,*) 'app=',app
            close(10)
c     Debug
            call intersection(0,dim,event,app,1,1,P1,u1,
     &           r(ir-1,itheta,iphi),r(ir,itheta,iphi),
     &           theta(itheta-1),theta(itheta),
     &           phi(iphi-1),phi(iphi),ca,cb,cc,1,1,
     &           solf1,surface1,Pint1,aerr1)
            if ((solf1.eq.1).and.(aerr1.eq.0)) then
               call copy_vector(dim,Pint1,P)
c     Debug
               call logfile_name(pindex,event,log_file)
               string='./nlines'
               call stringp_file_name(pindex,string,stringp_file)
               call get_nlines(log_file,stringp_file,nl)
               open(10,file=log_file(1:strlen(log_file)))
c     Debug
c               write(*,*) 'nl4=',nl
c     Debug
               do i=1,nl
                  read(10,*)
               enddo
c     Debug
c               write(*,*) 'ok4'
c     Debug
               if (nl.gt.0) then
                  write(10,*)
               endif
               write(10,*) 'Intersection found:'
               write(10,*) 'Pint1=',(Pint1(i),i=1,dim)
c     Debug
               close(10)
               goto 111
            else
               goto 222
            endif
c         else
c            write(10,*) 'maximal number of iterations reached'
c            write(10,*) 'process failed'
         endif
      else
c     Debug
         write(10,*) 'valid intersection found with cell:'
         write(10,*) 'Pint=',(Pint(i),i=1,dim)
c     Debug
         valid=1
         ir_out=ir
         itheta_out=itheta
         iphi_out=iphi
         call copy_vector(dim,P,P_out)
         call copy_vector(dim,u,u_out)
         s_out=surface
      endif
      
 666  continue

c     Debug
      if (valid.eq.1) then
         write(10,*) 'valid solution was found !'
         write(10,*) 'number of iterations:',niter
         close(10)
      else
         write(10,*) 'valid solution not found'
         close(10)
         call write_report(dim,event,Nr,Ntheta,Nphi,r,theta,phi,
     &        r_nodes,r_south_pole,r_north_pole,
     &        coeff_a,coeff_b,coeff_c,P_in,u_in)
      endif
c     Debug
      return
      end



      subroutine write_report(dim,event,Nr,Ntheta,Nphi,r,theta,phi,
     &     r_nodes,r_south_pole,r_north_pole,
     &     coeff_a,coeff_b,coeff_c,P_in,u_in)
      implicit none
      include 'max.inc'
      include 'mpif.h'
c     
c     Purpose: to write a report in a log file when the "last_chance"
c     procedure failed
c
c     Inputs
c       + dim: dimension of the vectors
c       + event: index of the event
c       + Nr,Ntheta,Nphi: number of cells
c       + r, theta, phi: limits of the spherical cells
c       + r_nodes: radius (altitude) at each one of the 4 nodes for
c                  each (theta,phi) angular sector
c       + r_south_pole: radius at south pole
c       + r_north_pole: radius at north pole
c       + coeff_a, coeff_b, coeff_c: a, b and c coefficients for each quarter
c       + P_in: (r,theta,phi) spherical coordinates of the entry point
c       + u_in: entry direction
c     Outputs
c       + update log file
c     
c     input
      integer dim,event
      integer Nr(1:Ntheta_mx,1:Nphi_mx),Ntheta,Nphi
      double precision r(0:Nz_mx,1:Ntheta_mx,1:Nphi_mx)
      double precision theta(0:Ntheta_mx)
      double precision phi(0:Nphi_mx)
      double precision r_nodes(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision r_south_pole,r_north_pole
      double precision coeff_a(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision coeff_b(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision coeff_c(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision P_in(1:Nmat_mx)
      double precision u_in(1:Nmat_mx)
c     MPICH
      integer ierr,errorcode,np,pindex
      integer stat(MPI_STATUS_SIZE)
c     MPICH
c     temp
      integer nl,i,solf,aerr,surface_int
      double precision Pint(1:Nmat_mx)
      integer cellf,ir,itheta,iphi,surface,app
      double precision n(1:Nmat_mx),prod
      double precision ca(1:4)
      double precision cb(1:4)
      double precision cc(1:4)
      character*100 log_file
      character*100 string
      character*100 stringp_file
      character*3 pindex_str3
      integer strlen
      character*100 label
c     MPICH
c     Get 'np', the number of processes
      call MPI_COMM_SIZE(MPI_COMM_WORLD,np,ierr)
c     Get 'pindex', the index of the current process
      call MPI_COMM_RANK(MPI_COMM_WORLD,pindex,ierr)
c     MPICH
      label='subroutine write_report'

      log_file='./logs/last_chance'
      call num2str3(pindex,pindex_str3)
      log_file=log_file(1:strlen(log_file))//
     &     pindex_str3(1:strlen(pindex_str3))//
     &     '.log'

      string='./nlines'
      call stringp_file_name(pindex,string,stringp_file)
      call get_nlines(log_file,stringp_file,nl)

      open(10,file=log_file(1:strlen(log_file)))
c     Debug
c      write(*,*) 'nl5=',nl
c     Debug
      do i=1,nl
         read(10,*)
      enddo
c     Debug
c      write(*,*) 'ok5'
c     Debug
      if (nl.gt.0) then
         write(10,*)
      endif
      write(10,*) 'routine last_chance failed for position:'
      write(10,*) 'P_in:',(P_in(i),i=1,dim)
      write(10,*) 'u_in:',(u_in(i),i=1,dim)
      write(10,*) 'event:',event
      close(10)

      call identify_valid_cell(dim,event,Nr,Ntheta,Nphi,r,theta,phi,
     &     r_nodes,r_south_pole,r_north_pole,
     &     coeff_a,coeff_b,coeff_c,P_in,u_in,
     &     cellf,ir,itheta,iphi,surface,n)

      string='./nlines'
      call stringp_file_name(pindex,string,stringp_file)
      call get_nlines(log_file,stringp_file,nl)
      open(10,file=log_file(1:strlen(log_file)))
c     Debug
c      write(*,*) 'nl6=',nl
c     Debug
      do i=1,nl
         read(10,*)
      enddo
c     Debug
c      write(*,*) 'ok6'
c     Debug
      if (nl.gt.0) then
         write(10,*)
      endif

      if (cellf.eq.0) then
         write(10,*) 'No valid cell could be identified'
      else
         write(10,*) 'position was identified in cell:'
         write(10,*) 'ir=',ir
         write(10,*) 'itheta=',itheta
         write(10,*) 'iphi=',iphi
         write(10,*) 'r=',r(ir-1,itheta,iphi),r(ir,itheta,iphi)
         write(10,*) 'theta=',theta(itheta-1),theta(itheta)
         write(10,*) 'phi=',phi(iphi-1),phi(iphi)
         write(10,*) 'input surface:',surface
         write(10,*) 'normal at input position:'
         write(10,*) 'n=',(n(i),i=1,dim)
         call scalar_product(dim,u_in,n,prod)
         write(10,*) 'scalar product=',prod
         call extract_coeffs(itheta,iphi,coeff_a,coeff_b,coeff_c,
     &        ca,cb,cc)
         app=0
         if (ir.eq.1) then
            call need2approximate(dim,Ntheta,r,theta,phi,
     &           r_nodes,r_south_pole,r_north_pole,itheta,iphi,app)
         endif
         write(10,*) 'app=',app
         write(10,*)'trying intersection'
         close(10)
         call intersection(0,dim,event,app,1,1,P_in,u_in,
     &        r(ir-1,itheta,iphi),r(ir,itheta,iphi),
     &        theta(itheta-1),theta(itheta),
     &        phi(iphi-1),phi(iphi),ca,cb,cc,1,1,
     &        solf,surface_int,Pint,aerr)

         string='./nlines'
         call stringp_file_name(pindex,string,stringp_file)
         call get_nlines(log_file,stringp_file,nl)
         open(10,file=log_file(1:strlen(log_file)))
c     Debug
c         write(*,*) 'nl7=',nl
c     Debug
         do i=1,nl
            read(10,*)
         enddo
c     Debug
c         write(*,*) 'ok7'
c     Debug
         if (nl.gt.0) then
            write(10,*)
         endif

         write(10,*) 'solf=',solf
         write(10,*) 'aerr=',aerr
         if (solf.eq.1) then
            write(*,*) 'surface_int=',surface_int
            write(*,*) 'Pint=',(Pint(i),i=1,dim)
         endif
      endif
      close(10)

      return
      end



      subroutine identify_valid_cell(dim,event,Nr,Ntheta,Nphi,
     &     r,theta,phi,
     &     r_nodes,r_south_pole,r_north_pole,
     &     coeff_a,coeff_b,coeff_c,P_in,u_in,
     &     cellf,ir_out,itheta_out,iphi_out,surface,n)
      implicit none
      include 'mpif.h'
      include 'max.inc'
c
c     Purpose: to identify a cell position where both "P_in" is on the surface,
c     and "u_in" is an input direction
c
c     Inputs
c       + dim: dimension of the vectors
c       + event: index of the event
c       + Nr,Ntheta,Nphi: number of cells
c       + r, theta, phi: limits of the spherical cells
c       + r_nodes: radius (altitude) at each one of the 4 nodes for
c                  each (theta,phi) angular sector
c       + r_south_pole: radius at south pole
c       + r_north_pole: radius at north pole
c       + coeff_a, coeff_b, coeff_c: a, b and c coefficients for each quarter
c       + P_in: (r,theta,phi) spherical coordinates of the entry point
c       + u_in: entry direction
c
c     Outputs:
c       + cellf: 1 if valid cell was found; 0 otherwise
c       + ir_out,itheta_out,iphi_out: index of the cell if cellf=1
c       + surface: index of the surface "P_in" is on
c       + n: normal direction at "P_in" position
c     

c     inputs
      integer dim,event
      integer Nr(1:Ntheta_mx,1:Nphi_mx),Ntheta,Nphi
      double precision r(0:Nz_mx,1:Ntheta_mx,1:Nphi_mx)
      double precision theta(0:Ntheta_mx)
      double precision phi(0:Nphi_mx)
      double precision r_nodes(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision r_south_pole,r_north_pole
      double precision coeff_a(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision coeff_b(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision coeff_c(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision P_in(1:Nmat_mx)
      double precision u_in(1:Nmat_mx)
c     outputs
      integer cellf
      integer ir_out,itheta_out,iphi_out
      integer surface
      double precision n(1:Nmat_mx)
c     MPICH
      integer ierr,errorcode,np,pindex
      integer stat(MPI_STATUS_SIZE)
c     MPICH
c     temp
      integer ir,itheta,iphi,app,inc,ong
      double precision pos(1:2),r_ground
      double precision ca(1:4)
      double precision cb(1:4)
      double precision cc(1:4)
      integer inb,on_surface,i
      double precision prod
      integer strlen
      character*100 log_file
      character*100 string
      character*100 stringp_file
      integer nl
      character*100 label
c     MPICH
c     Get 'np', the number of processes
      call MPI_COMM_SIZE(MPI_COMM_WORLD,np,ierr)
c     Get 'pindex', the index of the current process
      call MPI_COMM_RANK(MPI_COMM_WORLD,pindex,ierr)
c     MPICH
      label='subroutine identify_valid_cell'

      cellf=0
      do itheta=1,Ntheta
         do iphi=1,Nphi
            do ir=1,Nr(itheta,iphi)
               call extract_coeffs(itheta,iphi,coeff_a,coeff_b,coeff_c,
     &              ca,cb,cc)
               app=0
               if (ir.eq.1) then
                  call need2approximate(dim,Ntheta,r,theta,phi,
     &                 r_nodes,r_south_pole,r_north_pole,
     &                 itheta,iphi,app)
               endif
               inb=0
               if (app.eq.0) then
                  if ((P_in(1).ge.r(ir-1,itheta,iphi)).and.
     &                 (P_in(1).le.r(ir,itheta,iphi)).and.
     &                 (P_in(2).ge.theta(itheta-1)).and.
     &                 (P_in(2).le.theta(itheta)).and.
     &                 (P_in(3).ge.phi(iphi-1)).and.
     &                 (P_in(3).le.phi(iphi))) then
                     inb=1
                  endif
               else ! app=1
                  call is_in_column(P_in,theta(itheta-1),theta(itheta),
     &                 phi(iphi-1),phi(iphi),inc)
                  pos(1)=P_in(2)
                  pos(2)=P_in(3)
                  call ground_radius(pos,theta(itheta-1),theta(itheta),
     &                 phi(iphi-1),phi(iphi),ca,cb,cc,r_ground)
                  if ((inc.eq.1).and.(P_in(1).ge.r_ground-1.0D-4).and.
     &                 (P_in(1).le.r(ir,itheta,iphi))) then
                     inb=1
                  endif
               endif
               if (inb.eq.1) then
                  call cell_surface(dim,Nr,Ntheta,Nphi,r,theta,phi,
     &                 r_nodes,r_south_pole,r_north_pole,
     &                 coeff_a,coeff_b,coeff_c,P_in,ir,itheta,iphi,
     &                 on_surface,surface,n)
                  if (on_surface.eq.0) then

                     call logfile_name(pindex,event,log_file)
                     string='./nlines'
                     call stringp_file_name(pindex,string,stringp_file)
                     call get_nlines(log_file,stringp_file,nl)
                     open(10,file=log_file(1:strlen(log_file)))
c     Debug
c                     write(*,*) 'nl8=',nl
c     Debug
                     do i=1,nl
                        read(10,*)
                     enddo
c     Debug
c                     write(*,*) 'ok8'
c     Debug
                     if (nl.gt.0) then
                        write(10,*)
                     endif

                     write(10,*) 'Error from ',
     &                    label(1:strlen(label)),' :'
                     write(10,*) 'Position P_in:',(P_in(i),i=1,dim)
                     write(10,*) 'is not on the surface of cell:',
     &                    ir,itheta,iphi
                     write(10,*) 'r=',r(ir-1,itheta,iphi),
     &                    r(ir,itheta,iphi)
                     write(10,*) 'theta=',theta(itheta-1),
     &                    theta(itheta)
                     write(10,*) 'phi=',phi(iphi-1),
     &                    phi(iphi)
                     pos(1)=P_in(2)
                     pos(2)=P_in(3)
                     call extract_coeffs(itheta,iphi,
     &                    coeff_a,coeff_b,coeff_c,
     &                    ca,cb,cc)
                     call ground_radius(pos,
     &                    theta(itheta-1),theta(itheta),
     &                    phi(iphi-1),phi(iphi),
     &                    ca,cb,cc,r_ground)
                     write(10,*) 'r_ground=',r_ground

                     close(10)
                     cellf=0
                     goto 666
c                     stop
                  else
                     call scalar_product(dim,u_in,n,prod)
                     if (prod.gt.0.0D+0) then
                        cellf=1
                        ir_out=ir
                        itheta_out=itheta
                        iphi_out=iphi
                        goto 666
                     endif ! prod
                  endif ! on_surface
               endif ! inb

            enddo               ! iphi
         enddo                  ! itheta
      enddo                     ! iphi

 666  continue
      return
      end






      subroutine cell_surface(dim,Nr,Ntheta,Nphi,r,theta,phi,
     &     r_nodes,r_south_pole,r_north_pole,
     &     coeff_a,coeff_b,coeff_c,P_in,ir,itheta,iphi,
     &     ons,surface,n)
      implicit none
      include 'max.inc'
c
c     Purpose: to detect the index of the surface position "P" is on
c
c     Inputs
c       + dim: dimension of the vectors
c       + Nr,Ntheta,Nphi: number of cells
c       + r, theta, phi: limits of the spherical cells
c       + r_nodes: radius (altitude) at each one of the 4 nodes for
c                  each (theta,phi) angular sector
c       + r_south_pole: radius at south pole
c       + r_north_pole: radius at north pole
c       + coeff_a, coeff_b, coeff_c: a, b and c coefficients for each quarter
c       + P_in: (r,theta,phi) spherical coordinates of the entry point
c       + ir,itheta,iphi: index of the cell "P_in" has been detected on
c
c     Outputs:
c       + ons: 1 if particle is on cell surface; 0 otherwise
c       + surface: index (1-6) of the surface position "P" is on
c       + n: normal direction at position "P"
c

c     inputs
      integer dim
      integer Nr(1:Ntheta_mx,1:Nphi_mx),Ntheta,Nphi
      double precision r(0:Nz_mx,1:Ntheta_mx,1:Nphi_mx)
      double precision theta(0:Ntheta_mx)
      double precision phi(0:Nphi_mx)
      double precision r_nodes(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision r_south_pole,r_north_pole
      double precision coeff_a(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision coeff_b(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision coeff_c(1:Ntheta_mx,1:Nphi_mx,1:4)
      double precision P_in(1:Nmat_mx)
      integer ir,itheta,iphi
c     outputs
      integer ons,surface
      double precision n(1:Nmat_mx)
c     temp
      double precision pos(1:2),r_ground
      double precision ca(1:4)
      double precision cb(1:4)
      double precision cc(1:4)
      integer ong,app
      double precision epsilon_theta,epsilon_phi
      integer strlen
      character*100 label
      label='subroutine cell_surface'
      epsilon_theta=1.0D-5 ! rad
      epsilon_phi=1.0D-5 ! rad

      call extract_coeffs(itheta,iphi,coeff_a,coeff_b,coeff_c,
     &     ca,cb,cc)
      app=0
      if (ir.eq.1) then
         call need2approximate(dim,Ntheta,r,theta,phi,
     &        r_nodes,r_south_pole,r_north_pole,itheta,iphi,app)
      endif

c     now identify the surface the particle enters the cell from
      ons=0
      ong=0
c     Debug
c      write(*,*) 'now cell_surface'
c      write(*,*) 'app=',app
c     Debug
      if ((ir.eq.1).and.(app.eq.1)) then
c     test if particle is on the real ground (if possible)
         call is_on_ground(P_in,theta(itheta-1),theta(itheta),
     &        phi(iphi-1),phi(iphi),ca,cb,cc,ong)
c     Debug
c         write(*,*) 'ong=',ong
c     Debug
         if (ong.eq.1) then
            ons=1
            surface=1
            call ground_normal(dim,P_in,theta(itheta-1),theta(itheta),
     &           phi(iphi-1),phi(iphi),ca,cb,cc,n)
c     Debug
c            write(*,*) 'particle was found on the real ground'
c     Debug
c         else
c     Debug
c            write(*,*) 'particle is not on the real ground'
c     Debug
         endif
      endif
      if (ons.eq.0) then
         call is_on_surface(dim,r(ir-1,itheta,iphi),r(ir,itheta,iphi),
     &        theta(itheta-1),theta(itheta),
     &        phi(iphi-1),phi(iphi),
     &        P_in,ons,surface)
         call inside_normal(surface,P_in,n)
         if (ons.eq.1) then
c     Debug
c            write(*,*) 'particle was found on cell surface=',surface
c     Debug
         else
c     Debug
c            write(*,*) 'particle was not found on cell surface'
c     Debug
         endif
      endif
c     as a last resort, the particle may be on lateral surfaces,
c     between the real ground and r(0,itheta,iphi)
      if (ons.eq.0) then
         if (app.eq.1) then
            pos(1)=P_in(2)
            pos(2)=P_in(3)
            call ground_radius(pos,theta(itheta-1),theta(itheta),
     &           phi(iphi-1),phi(iphi),ca,cb,cc,r_ground)
c     Debug
c            write(*,*) 'r_ground=',r_ground
c     Debug
            if ((P_in(1).ge.r_ground-1.0D-4).and.
     &           (P_in(1).le.r(0,itheta,iphi))) then
               if (dabs(P_in(3)-phi(iphi-1)).lt.epsilon_phi) then
                  ons=1
                  surface=3
               else if (dabs(P_in(3)-phi(iphi)).lt.epsilon_phi) then
                  ons=1
                  surface=4
               else if (dabs(P_in(2)-theta(itheta-1)).lt.
     &                 epsilon_theta) then
                  ons=1
                  surface=5
               else if (dabs(P_in(2)-theta(itheta)).lt.epsilon_theta)
     &                 then
                  ons=1
                  surface=6
c               else
c                  write(*,*) 'error: theta=',P(2)
c                  write(*,*) 'and phi=',P(3)
c                  write(*,*) 'were not recognized'
c                  write(*,*) 'phi1=',phi(iphi-1)
c                  write(*,*) 'phi2=',phi(iphi)
c                  write(*,*) 'theta1=',theta(itheta-1)
c                  write(*,*) 'theta2=',theta(itheta)
               endif
               if (ons.eq.1) then
                  call inside_normal(surface,P_in,n)
c     Debug
c                  write(*,*) 'particle was found on lateral surface=',
c     &                 surface
c     Debug
               else
c     Debug
c                  write(*,*) 'part. was not found on lateral surface'
c     Debug
               endif
c            else
c               write(*,*) 'error: r=',P(1)
c               write(*,*) 'is not below r(0)=',r(0,itheta,iphi)
            endif
         else ! app=0
c     Debug
c            write(*,*) 'app=',app
c     Debug
         endif
      endif

      return
      end
