c     Copyright (C) 2008-2009 Vincent Eymet
c
c     Planet_EMC is free software; you can redistribute it and/or modify
c     it under the terms of the GNU General Public License as published by
c     the Free Software Foundation; either version 3, or (at your option)
c     any later version.

c     Planet_EMC is distributed in the hope that it will be useful,
c     but WITHOUT ANY WARRANTY; without even the implied warranty of
c     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c     GNU General Public License for more details.

c     You should have received a copy of the GNU General Public License
c     along with Planet_EMC; if not, see <http://www.gnu.org/licenses/>
c
      subroutine linear_system(db,n,a,y,valid,x)
      implicit none
      include 'max.inc'
c     
c     Purpose: to solve a Cramer linear system A*X=Y
c
c     Inputs:
c       + n: dimension of the system (must be greater than 1 !)
c       + a: (n*n) square matrix
c       + y: (n) vector of second-terms
c
c     Outputs:
c       + valid: 1 if a valid solution was found; 0 otherwise
c       + x: (n) vector of solutions if a valid solution was found
c

c     inputs
      integer n
      double precision a(1:Nmat_mx,1:Nmat_mx)
      double precision y(1:Nmat_mx)
c     outputs
      integer valid
      double precision x(1:Nmat_mx)
c     temp
      external determinant
      double precision a_tmp(1:Nmat_mx,1:Nmat_mx)
      double precision y_tmp(1:Nmat_mx)
      double precision det,eps
      double precision a11,b,c,sum
      double precision y_ref
      integer i,j,k,l,lif,li,iy
      integer strlen
      character*100 label
c     Debug
      integer db
c     Debug
      label='subroutine linear_system'

      if ((n.lt.1).or.(n.gt.Nmat_mx)) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Bad input parameter:'
         write(*,*) 'n=',n
         if (n.gt.Nmat_mx) then
            write(*,*) 'while Nmat_mx=',Nmat_mx
         endif
         stop
      endif
      valid=1

c     ------------------------------------------------------
c     case of dimension 1 system (1 equation only)
      if (n.eq.1) then
         if (a(1,1).eq.0.0D+0) then
            write(*,*) 'Error from ',
     &           label(1:strlen(label)),' :'
            write(*,*) 'a*x=y with a=',a(1,1)
            write(*,*) 'no solution !'
            valid=0
            goto 666
         else
            x(1)=y(1)/a(1,1)
            goto 666
         endif
      endif

c     ------------------------------------------------------
c     in the case all solutions y(i) are identical
      iy=1
      y_ref=y(1)
      do i=2,n
         if (y(i).ne.y_ref) then
            iy=0
            goto 123
         endif
      enddo ! i
 123  continue
      if (iy.eq.1) then
         do i=1,n-1
            x(i)=0.0D+0
         enddo
         x(n)=y_ref
         goto 666
      endif

c     ------------------------------------------------------
c     all other cases...

c     copy "a" onto "a_tmp" and "y" onto "y_tmp"(that will be modified)
      call copy_matrix(n,a,a_tmp)
      call copy_vector(n,y,y_tmp)

c     computing determinant
c     --------------------------------
c     modif 2009/04/29
c     
c     if your compiler supports a recursive compilation option (ifort uses the "-recursive" option)
c     then you can use subroutine "determinant"
c     Otherwise, you MUST use subroutine "det3"
c
      call det3(n,a_tmp,det)
c      call determinant(n,a_tmp,det,determinant)
c     modif 2009/04/29
c     --------------------------------
      if (det.lt.1.0D-10) then
         write(*,*) 'Error from ',
     &        label(1:strlen(label)),' :'
         write(*,*) 'determinant=',det
         write(*,*) 'matrix='
         do i=1,n
            write(*,*) (a(i,j),j=1,n),y(i)
         enddo ! i
         stop
      endif
      
      eps=1.0D-10
c     solving system using the Gauss method
      do i=2,n
 222     continue
         a11=a_tmp(i-1,i-1)
         if (dabs(a11).lt.eps) then
            lif=0
            do l=i,n
               if (a_tmp(i-1,l).ne.0.0D+0) then
                  li=l
                  lif=1
                  goto 111
               endif
            enddo ! l
 111        continue
            if (lif.eq.0) then
               write(*,*) 'Error from ',
     &              label(1:strlen(label)),' :'
               write(*,*) 'going down'
               write(*,*) 'a(',i-1,',',i-1,')=',a11
               write(*,*) 'and no switching line found'
               valid=0
               goto 666
            else
c     switch lines 'i-1' and 'li'
               call switch_lines(n,a_tmp,y_tmp,i-1,li)
               goto 222
            endif
         endif
         do j=i,n
            b=a_tmp(j,i-1)
            c=b/a11
            do k=i,n
               a_tmp(j,k)=a_tmp(j,k)-a_tmp(i-1,k)*c
            enddo ! k
            y_tmp(j)=y_tmp(j)-y_tmp(i-1)*c
            a_tmp(j,i-1)=0.0D+0
         enddo ! j
      enddo ! i

c     Debug
      if (db.eq.1) then
         write(*,*)
         write(*,*) 'a_tmp='
         do i=1,n
            write(*,*) (a_tmp(i,j),j=1,n),y_tmp(i)
         enddo
      endif
c     Debug

      do i=n,1,-1
         sum=0.0D+0
         if (i+1.le.n) then
            do j=n,i+1,-1
               sum=sum+a_tmp(i,j)*x(j)
            enddo ! j
         endif
         if (a_tmp(i,i).eq.0.0D+0) then
            write(*,*) 'Error from ',
     &           label(1:strlen(label)),' :'
            write(*,*) 'going up'
            write(*,*) 'a_tmp(',i,',',i,')=',a_tmp(i,i)
            valid=0
            goto 666
         else
            x(i)=(y_tmp(i)-sum)/a_tmp(i,i)
c     Debug
            if (db.eq.1) then
               write(*,*) 'x(',i,')=',x(i)
            endif
c     Debug
         endif
      enddo ! i

 666  continue
      return
      end



      subroutine switch_lines(n,a,y,i,j)
      implicit none
      include 'max.inc'
c
c     Purpose: to switch two lines on linear system parameters
c
c     Inputs:
c       + n: dimension of the system
c       + a: matrix "A" in "A.X=Y"
c       + y: vector "Y" in "A.X=Y"
c       + i: index of the first line
c       + li: index of the second line
c
c     Outputs:
c       + a: (updated) matrix "A" in "A.X=Y"
c       + y: (updated) vector "Y" in "A.X=Y"
c     

c     inputs
      integer n
      double precision a(1:Nmat_mx,1:Nmat_mx)
      double precision y(1:Nmat_mx)
      integer i,j
c     temp
      double precision line_tmp(1:Nmat_mx)
      double precision y_tmp
      integer k
      integer strlen
      character*100 label
      label='subroutine switch_lines'

      if ((n.lt.1).or.(n.gt.Nmat_mx)) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Bad input parameter:'
         write(*,*) 'n=',n
         if (n.gt.Nmat_mx) then
            write(*,*) 'while Nmat_mx=',Nmat_mx
         endif
         stop
      endif

      if ((i.lt.1).or.(i.gt.n)) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Bad input parameter:'
         write(*,*) 'i=',i
         stop
      endif

      if ((j.lt.1).or.(j.gt.n)) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Bad input parameter:'
         write(*,*) 'j=',j
         stop
      endif

      do k=1,n
         line_tmp(k)=a(j,k)
         a(j,k)=a(i,k)
      enddo
      y_tmp=y(j)
      y(j)=y(i)
      do k=1,n
         a(i,k)=line_tmp(k)
      enddo
      y(i)=y_tmp

      return
      end




      subroutine determinant(dim,mat,det,dummy_arg)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the determinant of a square matrix (of any size)
c     WARNING: this routine must not be used when you don't know whether or not
c     your compiler supports recursivity. g77 supports recursivity by default
c     (it does not need extra compilation options), ifort will compile and use this
c     routine, but execution results are WRONG if the program has not been
c     compiled using the "-recursive" option.
c
c     Inputs:
c       + dim: dimension of the matrix (>1)
c       + mat: dim*dim matrix
c     
c     Output:
c       + det: determinant
c       + dummy_arg: use "determinant" in the main call
c                  i.e. "call determinant(dim,mat,det,determinant)"
c                  This is a dummy argument, used for recursivity
c
c     input
      integer dim
      double precision mat(1:Nmat_mx,1:Nmat_mx)
c     output
      double precision det
c     temp
      external dummy_arg
      integer i,j
      integer row
      integer dim_tmp
      double precision sign
      double precision mat_tmp(1:Nmat_mx,1:Nmat_mx),det_tmp
      integer strlen
      character*100 label
      label='subroutine determinant'

c     Debug
c      write(*,*) 'mat='
c      do i=1,dim
c         write(*,*) (mat(i,j),j=1,dim)
c      enddo
c      write(*,*)
c     Debug

      if (dim.eq.3) then
         call det3(dim,mat,det)
      else
         dim_tmp=dim-1
         det=0.0D+0
         do row=1,dim
            sign=(-1.0D+0)**(row+1)
            if (row.gt.1) then
               do i=1,row-1
                  do j=2,dim
                     mat_tmp(j-1,i)=mat(j,i)
                  enddo         ! j
               enddo            ! i
            endif
            if (row.lt.dim) then
               do i=row+1,dim
                  do j=2,dim
                     mat_tmp(j-1,i-1)=mat(j,i)
                  enddo         ! j
               enddo            ! i
            endif
c     Debug
c     write(*,*) 'row=',row,' sign=',sign
c     do i=1,dim_tmp
c     write(*,*) (mat_tmp(i,j),j=1,dim_tmp)
c     enddo
c     write(*,*)
c     Debug
            call dummy_arg(dim_tmp,mat_tmp,det_tmp,dummy_arg)
            det=det+sign*mat(1,row)*det_tmp
         enddo                  ! row
      endif ! dim=3

      return
      end



      
      subroutine det3(dim,mat,det)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the determinant of a square 3*3 matrix
c
c     Inputs:
c       + dim: dimension of the matrix (MUST BE 3)
c       + mat: 3*3 matrix
c     
c     Output:
c       + det: determinant
c

c     input
      integer dim
      double precision mat(1:Nmat_mx,1:Nmat_mx)
c     output
      double precision det
c     temp
      integer strlen
      character*100 label
      label='subroutine det3'


      if (dim.ne.3) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Bad input argument:'
         write(*,*) 'dim=',dim
         stop
      endif

      det=mat(1,1)*(mat(2,2)*mat(3,3)-mat(2,3)*mat(3,2))
     &     -mat(1,2)*(mat(2,1)*mat(3,3)-mat(2,3)*mat(3,1))
     &     +mat(1,3)*(mat(2,1)*mat(3,2)-mat(2,2)*mat(3,1))

      return
      end




      subroutine cross_product_spher(n,u,v,cp)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the cross-product between two given vectors
c     in a spherical coordinates system
c
c     Inputs:
c       + n: size of vectors (MUST BE 3)
c       + u: (r,theta,phi) spherical coordinates of the first vector
c       + v: (r,theta,phi) spherical coordinates of the second vector
c     
c     Output:
c       + cp: (x,y,z) cartesian coordinates of the
c             cross-product between u and v (also a vector)
c

c     inputs
      integer n
      double precision u(1:Nmat_mx)
      double precision v(1:Nmat_mx)
c     output
      double precision cp(1:Nmat_mx)
c     temp
      double precision x1(1:Nmat_mx)
      double precision x2(1:Nmat_mx)
      integer strlen
      character*100 label
      label='subroutine cross_product_spher'

      if (n.ne.3) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Bad input argument:'
         write(*,*) 'n=',n
         stop
      endif

      call spher2cart(label,1,u(1),u(2),u(3),x1(1),x1(2),x1(3))
      call spher2cart(label,1,v(1),v(2),v(3),x2(1),x2(2),x2(3))

      call cross_product(n,x1,x2,cp)

      return
      end




      subroutine cross_product(n,u,v,ncp)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the cross-product between two given vectors
c     in a cartesian coordinates system
c
c     Inputs:
c       + n: size of vectors (MUST BE 3)
c       + u: (x,y,z) cartesian coordinates of the first vector
c       + v: (x,y,z) cartesian coordinates of the second vector
c     
c     Output:
c       + cp: (x,y,z) cartesian coordinates of the
c             cross-product between u and v (also a vector)
c

c     inputs
      integer n
      double precision u(1:Nmat_mx)
      double precision v(1:Nmat_mx)
c     output
      double precision ncp(1:Nmat_mx)
c     temp
      double precision cp(1:Nmat_mx)
      double precision nu(1:Nmat_mx)
      double precision nv(1:Nmat_mx)
      integer strlen
      character*100 label
      label='subroutine cross_product'

      if (n.ne.3) then
         write(*,*) 'Error from ',label(1:strlen(label)),' :'
         write(*,*) 'Bad input argument:'
         write(*,*) 'n=',n
         stop
      endif

      call normalize_vector(n,u,nu)
      call normalize_vector(n,v,nv)

      cp(1)=nu(2)*nv(3)-nu(3)*nv(2)
      cp(2)=nu(3)*nv(1)-nu(1)*nv(3)
      cp(3)=nu(1)*nv(2)-nu(2)*nv(1)
      call normalize_vector(n,cp,ncp)

      return
      end




      subroutine add_matrix(n,M1,M2,M)
      implicit none
      include 'max.inc'
c
c     Purpose: to add two square matrices
c
c     Inputs:
c       + n: size of the matrices
c       + M1: matrix number 1
c       + M2: matrix number 2
c     
c     Outputs:
c       + M=M1+M2 (also a square matrix of size n)
c
c     inputs
      integer n
      double precision M1(1:Nmat_mx,1:Nmat_mx)
      double precision M2(1:Nmat_mx,1:Nmat_mx)
c     outputs
      double precision M(1:Nmat_mx,1:Nmat_mx)
c     temp
      integer i,j

      do i=1,n
         do j=1,n
            M(i,j)=M1(i,j)+M2(i,j)
         enddo ! j
      enddo ! i

      return
      end



      subroutine multiply_matrix(n,M1,M2,M)
      implicit none
      include 'max.inc'
c
c     Purpose: to multiply two square matrices
c
c     Inputs:
c       + n: size of the matrices
c       + M1: matrix number 1
c       + M2: matrix number 2
c     
c     Outputs:
c       + M=M1*M2 (also a square matrix of size n)
c
c     inputs
      integer n
      double precision M1(1:Nmat_mx,1:Nmat_mx)
      double precision M2(1:Nmat_mx,1:Nmat_mx)
c     outputs
      double precision M(1:Nmat_mx,1:Nmat_mx)
c     temp
      integer i,j,k

      do i=1,n
         do j=1,n
            M(i,j)=0.0D+0
            do k=1,n
               M(i,j)=M(i,j)+M1(i,k)*M2(k,j)
            enddo ! k
         enddo ! j
      enddo ! i

      return
      end



      subroutine matrix_vector(n,M1,vector,M)
      implicit none
      include 'max.inc'
c
c     Purpose: to multiply a square matrix by a vector
c
c     Inputs:
c       + n: size of matrix M1 and vector
c       + M1: square matrix
c       + vector: vector of size n
c     
c     Outputs:
c       + M=M1*vector (also a vector of size n)
c
c     inputs
      integer n
      double precision M1(1:Nmat_mx,1:Nmat_mx)
      double precision vector(1:Nmat_mx)
c     outputs
      double precision M(1:Nmat_mx)
c     temp
      integer i,k

      do i=1,n
         M(i)=0.0D+0
         do k=1,n
            M(i)=M(i)+M1(i,k)*vector(k)
         enddo                  ! k
      enddo                     ! i

      return
      end



      subroutine matrix_scalar(n,M1,scalar,M)
      implicit none
      include 'max.inc'
c
c     Purpose: to multiply a square matrix by a real
c
c     Inputs:
c       + n: size of matrix M1 and vector
c       + M1: square matrix
c       + scalar: the real value the M1 matrix has to be multiplied by
c     
c     Outputs:
c       + M=M1*scalar (also a square matrix of size n)
c
c     inputs
      integer n
      double precision M1(1:Nmat_mx,1:Nmat_mx)
      double precision scalar
c     outputs
      double precision M(1:Nmat_mx,1:Nmat_mx)
c     temp
      integer i,j

      do i=1,n
         do j=1,n
            M(i,j)=M1(i,j)*scalar
         enddo                  ! j
      enddo                     ! i

      return
      end



      subroutine are_colinear(n,u,v,col)
      implicit none
      include 'max.inc'
c
c     Purpose: determine whether or not provided vectors "u" and "v"
c     are colinear
c
c     col: 0 if "u" and "v" are not colinear; 1 if they are
c
      integer n,col
      double precision u(1:Nmat_mx)
      double precision v(1:Nmat_mx)
      double precision prod,epsilon
      parameter(epsilon=1.0D-3)

      call scalar_product(n,u,v,prod)
      if (dabs(1.0D+0-dabs(prod)).lt.epsilon) then
         col=1
      else
         col=0
      endif

      return
      end



      subroutine scalar_product_spher(n,P1,P2,prod)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the scalar product between two positions (vectors)
c     defined on a spherical coordinates system
c
c     Inputs:
c       + n: dimension of vectors
c       + P1: (r1,theta1,phi1) spherical coordinates of the first position
c       + P2: (r2,theta2,phi2) spherical coordinates of the second position
c     
c     Output:
c       + prod: scalar product
c

c     inputs
      integer n
      double precision P1(1:Nmat_mx)
      double precision P2(1:Nmat_mx)
c     output
      double precision prod
c     temp
      double precision x1(1:Nmat_mx)
      double precision x2(1:Nmat_mx)
      integer strlen
      character*100 label
      label='subroutine scalar_product_spher'

      call spher2cart(label,1,P1(1),P1(2),P1(3),x1(1),x1(2),x1(3))
      call spher2cart(label,1,P2(1),P2(2),P2(3),x2(1),x2(2),x2(3))

      call scalar_product(n,x1,x2,prod)

      if (prod.gt.1.0D+0) then
         prod=1.0D+0
      else if (prod.lt.-1.0D+0) then
         prod=-1.0D+0
      endif

      return
      end


      


      subroutine scalar_product(n,u,v,prod)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the scalar product of two vectors
c     Input vectors u and v will be normalized so that the
c     resulting scalar product is between -1 and 1 (it can be
c     interpreted as the cosine of an angle between 0 and pi)
c
      integer n
      double precision u(1:Nmat_mx)
      double precision v(1:Nmat_mx)
      double precision nu(1:Nmat_mx)
      double precision nv(1:Nmat_mx)
      double precision prod

      call normalize_vector(n,u,nu)
      call normalize_vector(n,v,nv)
      call vector_vector(n,nu,nv,prod)

      if (prod.gt.1.0D+0) then
         prod=1.0D+0
      else if (prod.lt.-1.0D+0) then
         prod=-1.0D+0
      endif

      return
      end



      subroutine vector_vector(n,u,v,prod)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the product of two vectors
c
      integer n
      double precision u(1:Nmat_mx)
      double precision v(1:Nmat_mx)
      double precision prod
      integer i

      prod=0.0D+0
      do i=1,n
         prod=prod+u(i)*v(i)
      enddo

      return
      end



      subroutine normalize_vector(n,u,nu)
      implicit none
      include 'max.inc'
      integer n
      double precision u(1:Nmat_mx)
      double precision nu(1:Nmat_mx)
      double precision length
      integer i

      call vector_length(n,u,length)
      do i=1,n
         nu(i)=u(i)/length
      enddo

      return
      end



      subroutine add_vectors(n,u1,u2,v)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the sum of two vectors
c
      integer n
      double precision u1(1:Nmat_mx)
      double precision u2(1:Nmat_mx)
      double precision v(1:Nmat_mx)
      integer i

      do i=1,n
         v(i)=u1(i)+u2(i)
      enddo

      return
      end



      subroutine substract_vectors(n,u1,u2,v)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the sum of two vectors
c
      integer n
      double precision u1(1:Nmat_mx)
      double precision u2(1:Nmat_mx)
      double precision v(1:Nmat_mx)
      integer i

      do i=1,n
         v(i)=u1(i)-u2(i)
      enddo

      return
      end



      subroutine scalar_vector(n,scalar,u,v)
      implicit none
      include 'max.inc'
c
c     Purpose: to compute the product of a scalar and a vector
c
      integer n
      double precision scalar
      double precision u(1:Nmat_mx)
      double precision v(1:Nmat_mx)
      integer i

      do i=1,n
         v(i)=u(i)*scalar
      enddo

      return
      end


      subroutine copy_vector(n,u1,u2)
      implicit none
      include 'max.inc'
      integer n
      double precision u1(1:Nmat_mx)
      double precision u2(1:Nmat_mx)
      integer i

      do i=1,n
         u2(i)=u1(i)
      enddo

      return
      end


      subroutine copy_matrix(n,mat1,mat2)
      implicit none
      include 'max.inc'
      integer n
      double precision mat1(1:Nmat_mx,1:Nmat_mx)
      double precision mat2(1:Nmat_mx,1:Nmat_mx)
      integer i,j

      do i=1,n
         do j=1,n
            mat2(i,j)=mat1(i,j)
         enddo ! j
      enddo ! i

      return
      end


      subroutine identity_matrix(n,Id)
      implicit none
      include 'max.inc'
      integer n
      double precision Id(1:Nmat_mx,1:Nmat_mx)
      integer i,j

      do i=1,n
         do j=1,n
            if (i.ne.j) then
               Id(i,j)=0.0D+0
            else
               Id(i,j)=1.0D+0
            endif
         enddo ! j
      enddo ! i

      return
      end


      subroutine vector_length(n,u,length)
      implicit none
      include 'max.inc'
      integer n
      double precision u(1:Nmat_mx)
      double precision length
      
      call distance_cart(0.0D+0,0.0D+0,0.0D+0,u(1),u(2),u(3),length)

      return
      end
