c     Copyright (C) 2009 Vincent Eymet
c
c     "satellite" is free software; you can redistribute it and/or modify
c     it under the terms of the GNU General Public License as published by
c     the Free Software Foundation; either version 3, or (at your option)
c     any later version.

c     This software is distributed in the hope that it will be useful,
c     but WITHOUT ANY WARRANTY; without even the implied warranty of
c     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c     GNU General Public License for more details.

c     You should have received a copy of the GNU General Public License
c     along with the present software; if not, see <http://www.gnu.org/licenses/>
c

      subroutine find_free_process(free,nproc,fpf,proc)
      implicit none
c
c     Purpose: to find 'proc', the index of a free child process
c
c     Input:
c       + free: 0/1 value for each child process (0: busy process, 1: free process)
c       + nproc: number of child processes
c
c     Output:
c       + fpf: 0 if no free child process found; 1 otherwise
c       + proc: index of the free child process found, if fpf=1
c
      include 'max.inc'
      integer free(1:Nproc_mx)
      integer proc,nproc,fpf

      fpf=0
      do proc=1,nproc
         if (free(proc).eq.1) then
            fpf=1
            goto 123
         endif
      enddo
 123  continue

      return
      end



      subroutine stopped_processes(free,nproc,nsp)
      implicit none
c
c     Purpose: to identify the number of child processes that have been stopped
c
c     Input:
c       + free: array of free processes (a value of 2 means that the corresponding child process has been stopped)
c       + nproc: number of child processes
c     
c     Output:
c       + nsp: number of child processes that have been stopped
c     
      include 'max.inc'
      integer free(1:Nproc_mx)
      integer proc,nproc,nsp
      nsp=0
      do proc=1,nproc
         if (free(proc).eq.2) then
            nsp=nsp+1
         endif
      enddo

      return
      end


      subroutine set_Nchunks(Nint_sectors,Nchunks)
      implicit none
      
      integer Nint_sectors,Nchunks

      Nchunks=Nint_sectors

      return
      end


      
      subroutine send_global_data(dim,
     &     Ntheta,Nphi,Nb_gas,Nint_sectors,
     &     int_sectors,Radius,rspace,op_alpha,theta,phi,
     &     Psat,Pground,solar_flux_toa,solar_inc_flux_toa)
      implicit none
      include 'mpif.h'
      include 'max.inc'

c     MPICH
      integer ierr
      integer stat(MPI_STATUS_SIZE)
c     MPICH
      integer dim
      integer Nphi,Ntheta,Nb_gas,Nint_sectors
      integer int_sectors(1:Ntheta_mx*Nphi_mx,1:4)
      double precision Radius,rspace,op_alpha
      double precision theta(0:Ntheta_mx)
      double precision phi(0:Nphi_mx)
      double precision Psat(1:Nmat_mx)
      double precision Pground(1:Nmat_mx)
      double precision solar_flux_toa(1:Ntheta_mx,1:Nphi_mx,
     &     1:Ntheta_mx,1:Nphi_mx,Nb_gas_mx) ! (W/m)
      double precision solar_inc_flux_toa(1:Ntheta_mx,1:Nphi_mx,
     &     1:Ntheta_mx,1:Nphi_mx,Nb_gas_mx) ! (W/m)

      call MPI_BCAST(dim,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Ntheta,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Nphi,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Nb_gas,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Nint_sectors,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(int_sectors,Ntheta_mx*Nphi_mx*4,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Radius,1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(rspace,1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(op_alpha,1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(theta,Ntheta_mx+1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(phi,Nphi_mx+1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Psat,Nmat_mx,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Pground,Nmat_mx,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(solar_flux_toa,
     &     Ntheta_mx*Nphi_mx*Ntheta_mx*Nphi_mx*Nb_gas_mx,
     &     MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(solar_inc_flux_toa,
     &     Ntheta_mx*Nphi_mx*Ntheta_mx*Nphi_mx*Nb_gas_mx,
     &     MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)

      return
      end



      
      subroutine receive_global_data(dim,
     &     Ntheta,Nphi,Nb_gas,Nint_sectors,
     &     int_sectors,Radius,rspace,op_alpha,theta,phi,
     &     Psat,Pground,solar_flux_toa,solar_inc_flux_toa)
      implicit none
      include 'mpif.h'
      include 'max.inc'

c     MPICH
      integer ierr
      integer stat(MPI_STATUS_SIZE)
c     MPICH
      integer dim
      integer Nphi,Ntheta,Nb_gas,Nint_sectors
      integer int_sectors(1:Ntheta_mx*Nphi_mx,1:4)
      double precision Radius,rspace,op_alpha
      double precision theta(0:Ntheta_mx)
      double precision phi(0:Nphi_mx)
      double precision Psat(1:Nmat_mx)
      double precision Pground(1:Nmat_mx)
      double precision solar_flux_toa(1:Ntheta_mx,1:Nphi_mx,
     &     1:Ntheta_mx,1:Nphi_mx,Nb_gas_mx) ! (W/m)
      double precision solar_inc_flux_toa(1:Ntheta_mx,1:Nphi_mx,
     &     1:Ntheta_mx,1:Nphi_mx,Nb_gas_mx) ! (W/m)

      call MPI_BCAST(dim,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Ntheta,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Nphi,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Nb_gas,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Nint_sectors,1,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(int_sectors,Ntheta_mx*Nphi_mx*4,MPI_INTEGER
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Radius,1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(rspace,1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(op_alpha,1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(theta,Ntheta_mx+1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(phi,Nphi_mx+1,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Psat,Nmat_mx,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(Pground,Nmat_mx,MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(solar_flux_toa,
     &     Ntheta_mx*Nphi_mx*Ntheta_mx*Nphi_mx*Nb_gas_mx,
     &     MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)
      call MPI_BCAST(solar_inc_flux_toa,
     &     Ntheta_mx*Nphi_mx*Ntheta_mx*Nphi_mx*Nb_gas_mx,
     &     MPI_DOUBLE_PRECISION
     &     ,0,MPI_COMM_WORLD,ierr)

      return
      end



      subroutine send_partial_results(Pc_1s,Pc_tot_1s,
     &     inc_Pc_1s,inc_Pc_tot_1s)
      implicit none
      include 'mpif.h'
      include 'max.inc'

c     MPICH
      integer ierr
      integer stat(MPI_STATUS_SIZE)
c     MPICH
      double precision Pc_1s(1:Nb_gas_mx)
      double precision Pc_tot_1s
      double precision inc_Pc_1s(1:Nb_gas_mx)
      double precision inc_Pc_tot_1s

      call MPI_SEND(Pc_1s,Nb_gas_mx,
     &     MPI_DOUBLE_PRECISION,0,1,
     &     MPI_COMM_WORLD,ierr)
      call MPI_SEND(Pc_tot_1s,1,
     &     MPI_DOUBLE_PRECISION,0,1,
     &     MPI_COMM_WORLD,ierr)
      call MPI_SEND(inc_Pc_1s,Nb_gas_mx,
     &     MPI_DOUBLE_PRECISION,0,1,
     &     MPI_COMM_WORLD,ierr)
      call MPI_SEND(inc_Pc_tot_1s,1,
     &     MPI_DOUBLE_PRECISION,0,1,
     &     MPI_COMM_WORLD,ierr)
      
      return
      end



      subroutine receive_partial_results(proc,Pc_1s,Pc_tot_1s,
     &     inc_Pc_1s,inc_Pc_tot_1s)
      implicit none
      include 'mpif.h'
      include 'max.inc'

c     MPICH
      integer ierr
      integer stat(MPI_STATUS_SIZE)
c     MPICH
      integer proc
      double precision Pc_1s(1:Nb_gas_mx)
      double precision Pc_tot_1s
      double precision inc_Pc_1s(1:Nb_gas_mx)
      double precision inc_Pc_tot_1s

      call MPI_RECV(Pc_1s,Nb_gas_mx,
     &     MPI_DOUBLE_PRECISION,proc,
     &     MPI_ANY_TAG,MPI_COMM_WORLD,
     &     stat,ierr)
      call MPI_RECV(Pc_tot_1s,1,
     &     MPI_DOUBLE_PRECISION,proc,
     &     MPI_ANY_TAG,MPI_COMM_WORLD,
     &     stat,ierr)
      call MPI_RECV(inc_Pc_1s,Nb_gas_mx,
     &     MPI_DOUBLE_PRECISION,proc,
     &     MPI_ANY_TAG,MPI_COMM_WORLD,
     &     stat,ierr)
      call MPI_RECV(inc_Pc_tot_1s,1,
     &     MPI_DOUBLE_PRECISION,proc,
     &     MPI_ANY_TAG,MPI_COMM_WORLD,
     &     stat,ierr)
      
      return
      end
