c     Copyright (C) 2009 Vincent Eymet
c
c     "satellite" is free software; you can redistribute it and/or modify
c     it under the terms of the GNU General Public License as published by
c     the Free Software Foundation; either version 3, or (at your option)
c     any later version.

c     This software is distributed in the hope that it will be useful,
c     but WITHOUT ANY WARRANTY; without even the implied warranty of
c     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
c     GNU General Public License for more details.

c     You should have received a copy of the GNU General Public License
c     along with the present software; if not, see <http://www.gnu.org/licenses/>
c
      subroutine read_Ftoa(Ntheta,Nphi,Nb_gas,theta,phi,
     &     lambda_gas_lo,lambda_gas_hi,
     &     Radius,rspace,
     &     solar_flux_toa,
     &     solar_inc_flux_toa)
      implicit none
      include 'max.inc'
      include 'formats.inc'
c
c     Purpose: to read the flux distribution at the top of the atmosphere
c
c     Inputs: none
c
c     Outputs:
c       + Ntheta,Nphi: number of angular sectors over the latitude / over the longitude
c       + Nb_gas: number of gas narrowband spectral intervals
c       + theta: latitude angular discretization
c       + phi: longitude angular discretization
c       + lambda_gas_lo: lower limits of gas spectral narrowbands
c       + lambda_gas_hi: higher limits of gas spectral narrowbands
c       + Radius: planet radius in meters
c       + rspace: radius at space level in meters
c       + solar_flux_toa: outgoing solar flux (W/m) at the top of the atmosphere
c       Depends on 5 indexes: itheta,iphi,u_itheta,u_iphi,band
c         - (itheta,iphi): indexes of the angular sector the flux exits from
c         - (u_itheta,u_iphi): indexes of the angular sector the flux is considered in
c         - band: index of the gas narrowband spectral interval the flux is considered into
c       The flux dimension is W/m (watts per square meter of surface at space level).
c       + solar_inc_flux_toa: Monte-Carlo uncertainties over TOA outgoing fluxes (W/m)
c     
      integer Nphi,Ntheta,Nb_gas
      double precision theta(0:Ntheta_mx)
      double precision phi(0:Nphi_mx)
      double precision Radius,rspace
      double precision lambda_gas_lo(1:Nb_gas_mx)
      double precision lambda_gas_hi(1:Nb_gas_mx)
      double precision solar_flux_toa(1:Ntheta_mx,1:Nphi_mx,
     &     1:Ntheta_mx,1:Nphi_mx,Nb_gas_mx) ! (W/m)
      double precision solar_inc_flux_toa(1:Ntheta_mx,1:Nphi_mx,
     &     1:Ntheta_mx,1:Nphi_mx,Nb_gas_mx) ! (W/m)
      
      integer i,ios,itheta,iphi,band,u_itheta,u_iphi
      integer strlen
      character*100 Ftoa_file
      character*100 label
      label='subroutine read_Ftoa'

      Ftoa_file='./data/integrated_solar_TOA_fluxes.txt'

      open(14,file=Ftoa_file(1:strlen(Ftoa_file)),
     &     status='old',iostat=ios)
      if (ios.ne.0) then        ! file not found
         write(*,*) 'Error from read_data:'
         write(*,*) 'File could not be found:'
         write(*,*) Ftoa_file(1:strlen(Ftoa_file))
         stop
      endif

      write(*,*) 'Reading TOA fluxes from file:'
      write(*,*) Ftoa_file(1:strlen(Ftoa_file))

      do i=1,8
         read(14,*)
      enddo
      read(14,*) Ntheta
      read(14,*)
      read(14,*) Nphi
      read(14,*)
      read(14,50) (theta(itheta),itheta=0,Ntheta)
      read(14,*)
      read(14,50) (phi(iphi),iphi=0,Nphi)
      read(14,*)
      read(14,*) Nb_gas
      read(14,*)
      read(14,50) (lambda_gas_lo(band),band=1,Nb_gas)
      read(14,*)
      read(14,50) (lambda_gas_hi(band),band=1,Nb_gas)
      read(14,*)
      read(14,50) Radius
      read(14,*)
      read(14,50) rspace
      read(14,*)

c     Debug
c      do iphi=1,Nphi
c         do itheta=1,Ntheta
c            do u_iphi=1,Nphi
c               do u_itheta=1,Ntheta
c                  do band=1,Nb_gas
c                     solar_flux_toa(itheta,iphi,
c     &                    u_itheta,u_iphi,band)=1.0D+0
c                     solar_inc_flux_toa(itheta,iphi,
c     &                    u_itheta,u_iphi,band)=1.0D-2
c                  enddo ! band
c               enddo            ! u_itheta
c            enddo               ! u_iphi
c         enddo ! itheta
c      enddo ! iphi
c      write(*,*) 'warning from ',label(1:strlen(label)),' :'
c      write(*,*) 'TOA fluxes not read !'
c      goto 666
c     Debug

      do iphi=1,Nphi
         do itheta=1,Ntheta
            do u_iphi=1,Nphi
               do u_itheta=1,Ntheta
                  read(14,*) !iphi,itheta,u_iphi,u_itheta
                  read(14,50) (solar_flux_toa(itheta,iphi,
     &                 u_itheta,u_iphi,band),band=1,Nb_gas),
     &                 (solar_inc_flux_toa(itheta,iphi,
     &                 u_itheta,u_iphi,band),band=1,Nb_gas)
               enddo            ! u_itheta
            enddo               ! u_iphi
            do i=1,4
               read(14,*)
            enddo
         enddo ! itheta
      enddo ! iphi

 666  continue
      close(14)
      write(*,*) '...done'

      return
      end
